                                                                          /*  GENSPROC.SQL  */

--  Create the stored procedures that are used in generating sample data and in
--      running many of the labs.
--  Some of these procedures are not really part of the database, they are just 
--      used for loading the sample data into the database.  But some of them are
--      used throught the course.  For this reason, the is no script file that
--      drops these procedures.  The remain in the Credit database for the duration
--      of the course.
--  Don't forget to say 'USE Credit' first.

USE Credit
SET NOCOUNT ON
GO
IF ( (SELECT db_name()) <> 'Credit')
BEGIN
   RAISERROR ('Database "Credit" not found.',22,127) WITH LOG
END
GO


PRINT 'Begin GENSPROC.SQL'
GO



--  Create a procedure that will exclusively lock a table for the duration of a transaction

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.pto_locktable') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.pto_locktable
GO


CREATE PROCEDURE pto_locktable 
    @table_name sysname
AS
BEGIN
   EXECUTE ('DECLARE @scratch int IF EXISTS (SELECT * FROM ' + @table_name + ' (TABLOCKX HOLDLOCK)) BEGIN SELECT @scratch = @scratch END')
END
GO



--  Create inhouse randomization routines.

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.straight_rand') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.straight_rand
IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.skewed_rand') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.skewed_rand
GO



CREATE PROCEDURE straight_rand 
    @max_value int
AS
BEGIN
   RETURN (rand() * @max_value) + 1
END
GO



CREATE PROCEDURE skewed_rand 
    @max_value int
  , @skew_factor tinyint
AS
BEGIN
   DECLARE @rand_value int    SELECT @rand_value = @max_value
   WHILE (@skew_factor > 0)
   BEGIN
      SELECT @skew_factor = @skew_factor - 1
      EXEC @rand_value = straight_rand @rand_value
   END
   RETURN (@max_value - @rand_value) + 1
END
GO



--  Create a stored procedure to show the distribution of the data

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.show_dist') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.show_dist
GO

CREATE PROCEDURE show_dist 
    @table  sysname
  , @column sysname
AS
BEGIN
   CREATE TABLE #temp_table ( NUMBER_THAT_HAVE int NULL
                            , THIS_MANY        int NULL
                            )
   EXEC ("INSERT INTO #temp_table SELECT " + @column + ", COUNT(*) FROM " + @table + " GROUP BY " + @column)
   EXEC ("SELECT 'NUMBER THAT HAVE' = COUNT(*), THIS_MANY FROM #temp_table GROUP BY THIS_MANY ORDER BY THIS_MANY DESC") 
   DROP TABLE #temp_table
END
GO



--  Create procedures for loading members.  Create them in reverse dependency sequence.

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.load_members') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.load_members
IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.generate_firstname') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.generate_firstname
GO



CREATE PROCEDURE generate_firstname 
   @firstname shortstring OUTPUT 
AS
BEGIN
   DECLARE @limit int
   DECLARE @curr_iteration int
   SELECT @limit = round((rand() * 20) + 3, 0)
   SELECT @curr_iteration = 0
   SELECT @firstname = ''
   WHILE @curr_iteration < @limit
   BEGIN
      SELECT @firstname = @firstname + char(round((rand() * 25) + 1, 0) + 64)
      SELECT @curr_iteration = @curr_iteration + 1
   END
   IF SUBSTRING(@firstname,1,1) = ' '
   BEGIN
      SELECT @firstname = SUBSTRING(@firstname,2,16)
   END
END
GO



CREATE PROCEDURE load_members 
   @target_member_count int 
AS
BEGIN
   DECLARE @member_region_no numeric_id
   DECLARE @member_corp_no numeric_id
   DECLARE @member_lastname shortstring
   DECLARE @member_firstname shortstring

   DECLARE @no_of_regions int      SELECT @no_of_regions = COUNT(*) FROM region
   DECLARE @no_of_corps int        SELECT @no_of_corps = COUNT(*) FROM corporation
   DECLARE @no_of_corps_t5 int     SELECT @no_of_corps_t5 = @no_of_corps * 2

   DECLARE tnames_cursor CURSOR FOR SELECT lastname FROM lastname_table

   WHILE @target_member_count > 0
   BEGIN
      BEGIN TRANSACTION

      OPEN tnames_cursor
      WHILE (@@FETCH_STATUS = @@FETCH_STATUS)
      BEGIN
         FETCH NEXT FROM tnames_cursor INTO @member_lastname
         IF (@@FETCH_STATUS <> 0)
            BREAK
         EXEC @member_region_no = skewed_rand @no_of_regions, 1
         EXEC @member_corp_no = skewed_rand @no_of_corps_t5, 2
         IF (@member_corp_no > @no_of_corps) SELECT @member_corp_no = NULL
         EXEC generate_firstname @member_firstname OUTPUT
         EXEC addmember
             @region_no        = @member_region_no
           , @corp_no          = @member_corp_no
           , @lastname         = @member_lastname
           , @firstname        = @member_firstname
           , @middleinitial    = ' '
           , @street           = '  '
           , @city             = '  '
           , @state_prov       = '  '
           , @country          = '  '
           , @mail_code        = '  '
           , @phone_no         = '  '
         SELECT @target_member_count = @target_member_count - 1
      END
      CLOSE tnames_cursor

      COMMIT TRANSACTION
      BACKUP LOG Credit WITH TRUNCATE_ONLY
   END
   DEALLOCATE tnames_cursor
END
GO



--  Create procedure for loading charges.

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.load_charges') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.load_charges
GO

CREATE PROCEDURE load_charges 
    @target_charge_count int
AS
BEGIN
   DECLARE @charge_category_no int
   DECLARE @charge_provider_no int
   DECLARE @charge_member_no int
   DECLARE @charge_charge_dt datetime
   DECLARE @charge_charge_amt money

   DECLARE @no_of_members int       SELECT @no_of_members = COUNT(*) FROM member
   DECLARE @no_of_providers int     SELECT @no_of_providers = COUNT(*) FROM provider
   DECLARE @no_of_categories int    SELECT @no_of_categories = COUNT(*) FROM category

   DECLARE @tran_size int           SELECT @tran_size = 4000

   BEGIN TRANSACTION
   WHILE @target_charge_count > 0
   BEGIN
      SELECT @target_charge_count = @target_charge_count - 1
      IF (@target_charge_count % @tran_size = 0) 
      BEGIN
         COMMIT TRANSACTION
         BACKUP LOG Credit WITH TRUNCATE_ONLY
         BEGIN TRANSACTION
      END

      EXEC @charge_category_no = skewed_rand @no_of_categories, 1
      EXEC @charge_provider_no = skewed_rand @no_of_providers, 3
      EXEC @charge_member_no   = skewed_rand @no_of_members, 2
      SELECT @charge_charge_dt = dateadd(day, -(rand()*120), getdate())
      SELECT @charge_charge_amt = convert(money, round((rand() * 5000 + 1), 0))

      EXEC inputcharge
          @member_no        = @charge_member_no   
        , @provider_no      = @charge_provider_no
        , @category_no      = @charge_category_no
        , @charge_dt        = @charge_charge_dt
        , @charge_amt       = @charge_charge_amt
   END
   COMMIT TRANSACTION
END
GO



--  Create procedure for loading statements.

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.load_statements') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.load_statements
GO

CREATE PROCEDURE load_statements 
   @statement_dt datetime
AS
BEGIN
   DECLARE @batch_size int                 SELECT @batch_size = 100
   DECLARE @starting_member_no numeric_id  SELECT @starting_member_no = 1
   DECLARE @ending_member_no numeric_id    SELECT @ending_member_no = COUNT(*) FROM member
   DECLARE @member_no numeric_id           SELECT @member_no = @starting_member_no - 1

   SELECT @statement_dt = CONVERT(CHAR(12),@statement_dt,9)               --  Trim off the time.

   BEGIN TRANSACTION
   WHILE @member_no < @ending_member_no
   BEGIN
      SELECT @member_no = @member_no + 1

      IF @member_no % @batch_size = 0 
      BEGIN
         COMMIT TRANSACTION
         BACKUP LOG Credit WITH TRUNCATE_ONLY
         BEGIN TRANSACTION
      END

      EXEC generatestatement @member_no, @statement_dt
   END
   COMMIT TRANSACTION
   BACKUP LOG Credit WITH TRUNCATE_ONLY
END
GO



--  Create procedure for loading payments.

IF EXISTS (SELECT * FROM sysobjects WHERE id = object_id('dbo.load_payments') and sysstat & 0xf = 4)
   DROP PROCEDURE dbo.load_payments
GO

CREATE PROCEDURE load_payments 
    @starting_dt datetime
  , @ending_dt datetime
AS
BEGIN
   DECLARE @batch_size int                   SELECT @batch_size = 1000

   DECLARE @member_no numeric_id
   DECLARE @due_dt datetime
   DECLARE @statement_amt     money
   DECLARE @statement_current money
   DECLARE @statement_pastdue money
   DECLARE @statement_latechg money

   DECLARE @payment_principle money

   DECLARE @highest_member_no numeric_id     SELECT @highest_member_no = COUNT(*) FROM member

   DECLARE @starting_member_no numeric_id  SELECT @starting_member_no = 1
   DECLARE @ending_member_no numeric_id    SELECT @ending_member_no = @batch_size

   SELECT @starting_dt = CONVERT(CHAR(12),@starting_dt,9)
   SELECT @ending_dt = CONVERT(CHAR(12),@ending_dt,9)    

   DECLARE statements_cursor CURSOR FOR 
      SELECT member_no, due_dt, statement_amt, statement_current, statement_pastdue, statement_latechg
      FROM statement
      WHERE member_no BETWEEN @starting_member_no AND @ending_member_no
        AND statement_dt > @starting_dt AND statement_dt <= @ending_dt 
        AND statement_amt > 0

   WHILE @starting_member_no <= @highest_member_no
   BEGIN
      BEGIN TRANSACTION
      OPEN statements_cursor

      WHILE (@@FETCH_STATUS = @@FETCH_STATUS)
      BEGIN
         FETCH NEXT FROM statements_cursor INTO @member_no, @due_dt, @statement_amt, @statement_current, @statement_pastdue, @statement_latechg
         IF (@@FETCH_STATUS <> 0)
            BREAK
         SELECT @payment_principle = @statement_current + @statement_pastdue
         EXEC inputpayment @member_no, @due_dt, @statement_amt, @payment_principle, @statement_latechg
      END

      CLOSE statements_cursor
      COMMIT TRANSACTION
      BACKUP LOG Credit WITH NO_LOG

      SELECT @starting_member_no = @ending_member_no + 1
      SELECT @ending_member_no = @ending_member_no + @batch_size
   END

   DEALLOCATE statements_cursor
END
GO



